@@ -3,18 +3,20 @@
 #
 # License: BSD 3 clause
 from __future__ import print_function
-import locale
 import numpy as np
 import numba
 
 from pynndescent.utils import norm, tau_rand
-from pynndescent.distances import kantorovich
-
-locale.setlocale(locale.LC_NUMERIC, "C")
+from pynndescent.distances import (
+    kantorovich,
+    jensen_shannon_divergence,
+    symmetric_kl_divergence,
+)
 
 FLOAT32_EPS = np.finfo(np.float32).eps
 FLOAT32_MAX = np.finfo(np.float32).max
 
+
 # Just reproduce a simpler version of numpy isclose (not numba supported yet)
 @numba.njit(cache=True)
 def isclose(a, b, rtol=1.0e-5, atol=1.0e-8):
@@ -50,6 +52,61 @@ def arr_intersect(ar1, ar2):
     return aux[:-1][aux[1:] == aux[:-1]]
 
 
+# Some things require size of intersection; do this quickly; assume sorted arrays for speed
+@numba.njit(
+    [
+        "i4(i4[:],i4[:])",
+        numba.types.int32(
+            numba.types.Array(numba.types.int32, 1, "C", readonly=True),
+            numba.types.Array(numba.types.int32, 1, "C", readonly=True),
+        ),
+    ],
+    locals={
+        "i1": numba.uint16,
+        "i2": numba.uint16,
+    },
+)
+def fast_intersection_size(ar1, ar2):
+    if ar1.shape[0] == 0 or ar2.shape[0] == 0:
+        return 0
+
+    # NOTE: We assume arrays are sorted; if they are not this will break
+    i1 = 0
+    i2 = 0
+    limit1 = ar1.shape[0] - 1
+    limit2 = ar2.shape[0] - 1
+    j1 = ar1[i1]
+    j2 = ar2[i2]
+
+    result = 0
+
+    while True:
+        if j1 == j2:
+            result += 1
+            if i1 < limit1:
+                i1 += 1
+                j1 = ar1[i1]
+            else:
+                break
+
+            if i2 < limit2:
+                i2 += 1
+                j2 = ar2[i2]
+            else:
+                break
+
+        elif j1 < j2 and i1 < limit1:
+            i1 += 1
+            j1 = ar1[i1]
+        elif j2 < j1 and i2 < limit2:
+            i2 += 1
+            j2 = ar2[i2]
+        else:
+            break
+
+    return result
+
+
 @numba.njit(
     [
         numba.types.Tuple(
@@ -196,7 +253,125 @@ def sparse_mul(ind1, data1, ind2, data2):
     return result_ind, result_data
 
 
-@numba.njit(cache=True)
+@numba.njit(
+    [
+        # "Tuple((i4[::1],f4[::1]))(i4[::1],f4[::1],i4[::1],f4[::1])",
+        numba.types.float32(
+            numba.types.Array(numba.types.int32, 1, "C", readonly=True),
+            numba.types.Array(numba.types.float32, 1, "C", readonly=True),
+            numba.types.Array(numba.types.int32, 1, "C", readonly=True),
+            numba.types.Array(numba.types.float32, 1, "C", readonly=True),
+        )
+    ],
+    fastmath=True,
+    locals={
+        "result": numba.types.float32,
+        "val": numba.types.float32,
+        "i1": numba.types.uint16,
+        "i2": numba.types.uint16,
+        "j1": numba.types.int32,
+        "j2": numba.types.int32,
+    },
+    cache=True,
+)
+def sparse_dot_product(ind1, data1, ind2, data2):
+    dim1 = ind1.shape[0]
+    dim2 = ind2.shape[0]
+
+    result = 0.0
+
+    i1 = 0
+    i2 = 0
+    j1 = ind1[i1]
+    j2 = ind2[i2]
+
+    # pass through both index lists
+    while True:
+        if j1 == j2:
+            val = data1[i1] * data2[i2]
+            result += val
+            i1 += 1
+            if i1 >= dim1:
+                return result
+            j1 = ind1[i1]
+            i2 += 1
+            if i2 >= dim2:
+                return result
+            j2 = ind2[i2]
+        elif j1 < j2:
+            i1 += 1
+            if i1 >= dim1:
+                return result
+            j1 = ind1[i1]
+        else:
+            i2 += 1
+            if i2 >= dim2:
+                return result
+            j2 = ind2[i2]
+
+    return result  # unreachable
+
+
+# Return dense vectors supported on the union of the non-zero valued indices
+@numba.njit()
+def dense_union(ind1, data1, ind2, data2):
+    result_ind = arr_union(ind1, ind2)
+    result_data1 = np.zeros(result_ind.shape[0], dtype=np.float32)
+    result_data2 = np.zeros(result_ind.shape[0], dtype=np.float32)
+
+    i1 = 0
+    i2 = 0
+    nnz = 0
+
+    # pass through both index lists
+    while i1 < ind1.shape[0] and i2 < ind2.shape[0]:
+        j1 = ind1[i1]
+        j2 = ind2[i2]
+
+        if j1 == j2:
+            val = data1[i1] + data2[i2]
+            if val != 0:
+                result_data1[nnz] = data1[i1]
+                result_data2[nnz] = data2[i2]
+                nnz += 1
+            i1 += 1
+            i2 += 1
+        elif j1 < j2:
+            val = data1[i1]
+            if val != 0:
+                result_data1[nnz] = data1[i1]
+                nnz += 1
+            i1 += 1
+        else:
+            val = data2[i2]
+            if val != 0:
+                result_data2[nnz] = data2[i2]
+                nnz += 1
+            i2 += 1
+
+    # pass over the tails
+    while i1 < ind1.shape[0]:
+        val = data1[i1]
+        if val != 0:
+            result_data1[nnz] = data1[i1]
+            nnz += 1
+        i1 += 1
+
+    while i2 < ind2.shape[0]:
+        val = data2[i2]
+        if val != 0:
+            result_data2[nnz] = data2[i2]
+            nnz += 1
+        i2 += 1
+
+    # truncate to the correct length in case there were zeros
+    result_data1 = result_data1[:nnz]
+    result_data2 = result_data2[:nnz]
+
+    return result_data1, result_data2
+
+
+@numba.njit(fastmath=True)
 def sparse_euclidean(ind1, data1, ind2, data2):
     _, aux_data = sparse_diff(ind1, data1, ind2, data2)
     result = 0.0
@@ -223,7 +398,6 @@ def sparse_euclidean(ind1, data1, ind2, data2):
         "dim": numba.types.intp,
         "i": numba.types.uint16,
     },
-    cache=True,
 )
 def sparse_squared_euclidean(ind1, data1, ind2, data2):
     _, aux_data = sparse_diff(ind1, data1, ind2, data2)
@@ -234,7 +408,7 @@ def sparse_squared_euclidean(ind1, data1, ind2, data2):
     return result
 
 
-@numba.njit(cache=True)
+@numba.njit()
 def sparse_manhattan(ind1, data1, ind2, data2):
     _, aux_data = sparse_diff(ind1, data1, ind2, data2)
     result = 0.0
@@ -243,7 +417,7 @@ def sparse_manhattan(ind1, data1, ind2, data2):
     return result
 
 
-@numba.njit(cache=True)
+@numba.njit()
 def sparse_chebyshev(ind1, data1, ind2, data2):
     _, aux_data = sparse_diff(ind1, data1, ind2, data2)
     result = 0.0
@@ -252,7 +426,7 @@ def sparse_chebyshev(ind1, data1, ind2, data2):
     return result
 
 
-@numba.njit(cache=True)
+@numba.njit()
 def sparse_minkowski(ind1, data1, ind2, data2, p=2.0):
     _, aux_data = sparse_diff(ind1, data1, ind2, data2)
     result = 0.0
@@ -261,13 +435,13 @@ def sparse_minkowski(ind1, data1, ind2, data2, p=2.0):
     return result ** (1.0 / p)
 
 
-@numba.njit(cache=True)
+@numba.njit()
 def sparse_hamming(ind1, data1, ind2, data2, n_features):
     num_not_equal = sparse_diff(ind1, data1, ind2, data2)[0].shape[0]
     return float(num_not_equal) / n_features
 
 
-@numba.njit(cache=True)
+@numba.njit()
 def sparse_canberra(ind1, data1, ind2, data2):
     abs_data1 = np.abs(data1)
     abs_data2 = np.abs(data2)
@@ -295,7 +469,6 @@ def sparse_canberra(ind1, data1, ind2, data2):
         ),
     ],
     fastmath=True,
-    cache=True,
 )
 def sparse_bray_curtis(ind1, data1, ind2, data2):  # pragma: no cover
     _, denom_data = sparse_sum(ind1, data1, ind2, data2)
@@ -317,10 +490,10 @@ def sparse_bray_curtis(ind1, data1, ind2, data2):  # pragma: no cover
     return float(numerator) / denominator
 
 
-@numba.njit(cache=True)
+@numba.njit()
 def sparse_jaccard(ind1, data1, ind2, data2):
-    num_non_zero = arr_union(ind1, ind2).shape[0]
-    num_equal = arr_intersect(ind1, ind2).shape[0]
+    num_equal = fast_intersection_size(ind1, ind2)
+    num_non_zero = ind1.shape[0] + ind2.shape[0] - num_equal
 
     if num_non_zero == 0:
         return 0.0
@@ -340,36 +513,39 @@ def sparse_jaccard(ind1, data1, ind2, data2):
     ],
     fastmath=True,
     locals={"num_non_zero": numba.types.intp, "num_equal": numba.types.intp},
-    cache=True,
 )
 def sparse_alternative_jaccard(ind1, data1, ind2, data2):
-    num_non_zero = arr_union(ind1, ind2).shape[0]
-    num_equal = arr_intersect(ind1, ind2).shape[0]
+    num_equal = fast_intersection_size(ind1, ind2)
+    num_non_zero = ind1.shape[0] + ind2.shape[0] - num_equal
 
     if num_non_zero == 0:
         return 0.0
+    elif num_equal == 0:
+        return FLOAT32_MAX
     else:
         return -np.log2(num_equal / num_non_zero)
+        # return (num_non_zero - num_equal) / num_equal
 
 
 @numba.vectorize(fastmath=True)
 def correct_alternative_jaccard(v):
     return 1.0 - pow(2.0, -v)
+    # return v / (v + 1)
 
 
-@numba.njit(cache=True)
+@numba.njit()
 def sparse_matching(ind1, data1, ind2, data2, n_features):
-    num_true_true = arr_intersect(ind1, ind2).shape[0]
-    num_non_zero = arr_union(ind1, ind2).shape[0]
+    num_true_true = fast_intersection_size(ind1, ind2)
+    num_non_zero = ind1.shape[0] + ind2.shape[0] - num_true_true
     num_not_equal = num_non_zero - num_true_true
 
     return float(num_not_equal) / n_features
 
 
-@numba.njit(cache=True)
+@numba.njit()
 def sparse_dice(ind1, data1, ind2, data2):
-    num_true_true = arr_intersect(ind1, ind2).shape[0]
-    num_non_zero = arr_union(ind1, ind2).shape[0]
+    num_true_true = fast_intersection_size(ind1, ind2)
+    num_non_zero = ind1.shape[0] + ind2.shape[0] - num_true_true
     num_not_equal = num_non_zero - num_true_true
 
     if num_not_equal == 0.0:
@@ -378,10 +554,10 @@ def sparse_dice(ind1, data1, ind2, data2):
         return num_not_equal / (2.0 * num_true_true + num_not_equal)
 
 
-@numba.njit(cache=True)
+@numba.njit()
 def sparse_kulsinski(ind1, data1, ind2, data2, n_features):
-    num_true_true = arr_intersect(ind1, ind2).shape[0]
-    num_non_zero = arr_union(ind1, ind2).shape[0]
+    num_true_true = fast_intersection_size(ind1, ind2)
+    num_non_zero = ind1.shape[0] + ind2.shape[0] - num_true_true
     num_not_equal = num_non_zero - num_true_true
 
     if num_not_equal == 0:
@@ -392,21 +568,21 @@ def sparse_kulsinski(ind1, data1, ind2, data2, n_features):
         )
 
 
-@numba.njit(cache=True)
+@numba.njit()
 def sparse_rogers_tanimoto(ind1, data1, ind2, data2, n_features):
-    num_true_true = arr_intersect(ind1, ind2).shape[0]
-    num_non_zero = arr_union(ind1, ind2).shape[0]
+    num_true_true = fast_intersection_size(ind1, ind2)
+    num_non_zero = ind1.shape[0] + ind2.shape[0] - num_true_true
     num_not_equal = num_non_zero - num_true_true
 
     return (2.0 * num_not_equal) / (n_features + num_not_equal)
 
 
-@numba.njit(cache=True)
+@numba.njit()
 def sparse_russellrao(ind1, data1, ind2, data2, n_features):
     if ind1.shape[0] == ind2.shape[0] and np.all(ind1 == ind2):
         return 0.0
 
-    num_true_true = arr_intersect(ind1, ind2).shape[0]
+    num_true_true = fast_intersection_size(ind1, ind2)
 
     if num_true_true == np.sum(data1 != 0) and num_true_true == np.sum(data2 != 0):
         return 0.0
@@ -414,19 +590,19 @@ def sparse_russellrao(ind1, data1, ind2, data2, n_features):
         return float(n_features - num_true_true) / (n_features)
 
 
-@numba.njit(cache=True)
+@numba.njit()
 def sparse_sokal_michener(ind1, data1, ind2, data2, n_features):
-    num_true_true = arr_intersect(ind1, ind2).shape[0]
-    num_non_zero = arr_union(ind1, ind2).shape[0]
+    num_true_true = fast_intersection_size(ind1, ind2)
+    num_non_zero = ind1.shape[0] + ind2.shape[0] - num_true_true
     num_not_equal = num_non_zero - num_true_true
 
     return (2.0 * num_not_equal) / (n_features + num_not_equal)
 
 
-@numba.njit(cache=True)
+@numba.njit()
 def sparse_sokal_sneath(ind1, data1, ind2, data2):
-    num_true_true = arr_intersect(ind1, ind2).shape[0]
-    num_non_zero = arr_union(ind1, ind2).shape[0]
+    num_true_true = fast_intersection_size(ind1, ind2)
+    num_non_zero = ind1.shape[0] + ind2.shape[0] - num_true_true
     num_not_equal = num_non_zero - num_true_true
 
     if num_not_equal == 0.0:
@@ -435,7 +611,7 @@ def sparse_sokal_sneath(ind1, data1, ind2, data2):
         return num_not_equal / (0.5 * num_true_true + num_not_equal)
 
 
-@numba.njit(cache=True)
+@numba.njit()
 def sparse_cosine(ind1, data1, ind2, data2):
     _, aux_data = sparse_mul(ind1, data1, ind2, data2)
     result = 0.0
@@ -463,7 +639,6 @@ def sparse_cosine(ind1, data1, ind2, data2):
         "dim": numba.types.intp,
         "i": numba.types.uint16,
     },
-    cache=True,
 )
 def sparse_alternative_cosine(ind1, data1, ind2, data2):
     _, aux_data = sparse_mul(ind1, data1, ind2, data2)
@@ -492,13 +667,9 @@ def sparse_correct_alternative_cosine(d):
         return 1.0 - pow(2.0, -d)
 
 
-@numba.njit(cache=True)
+@numba.njit()
 def sparse_dot(ind1, data1, ind2, data2):
-    _, aux_data = sparse_mul(ind1, data1, ind2, data2)
-    result = 0.0
-
-    for val in aux_data:
-        result += val
+    result = sparse_dot_product(ind1, data1, ind2, data2)
 
     return 1.0 - result
 
@@ -508,17 +679,10 @@ def sparse_dot(ind1, data1, ind2, data2):
     fastmath=True,
     locals={
         "result": numba.types.float32,
-        "dim": numba.types.intp,
-        "i": numba.types.uint16,
     },
-    cache=True,
 )
 def sparse_alternative_dot(ind1, data1, ind2, data2):
-    _, aux_data = sparse_mul(ind1, data1, ind2, data2)
-    result = 0.0
-    dim = len(aux_data)
-    for i in range(dim):
-        result += aux_data[i]
+    result = sparse_dot_product(ind1, data1, ind2, data2)
 
     if result <= 0.0:
         return FLOAT32_MAX
@@ -526,7 +690,7 @@ def sparse_alternative_dot(ind1, data1, ind2, data2):
         return -np.log2(result)
 
 
-@numba.njit(cache=True)
+@numba.njit()
 def sparse_correlation(ind1, data1, ind2, data2, n_features):
 
     mu_x = 0.0
@@ -555,10 +719,10 @@ def sparse_correlation(ind1, data1, ind2, data2, n_features):
         shifted_data2[i] = data2[i] - mu_y
 
     norm1 = np.sqrt(
-        (norm(shifted_data1) ** 2) + (n_features - ind1.shape[0]) * (mu_x ** 2)
+        (norm(shifted_data1) ** 2) + (n_features - ind1.shape[0]) * (mu_x**2)
     )
     norm2 = np.sqrt(
-        (norm(shifted_data2) ** 2) + (n_features - ind2.shape[0]) * (mu_y ** 2)
+        (norm(shifted_data2) ** 2) + (n_features - ind2.shape[0]) * (mu_y**2)
     )
 
     dot_prod_inds, dot_prod_data = sparse_mul(ind1, shifted_data1, ind2, shifted_data2)
@@ -587,7 +751,7 @@ def sparse_correlation(ind1, data1, ind2, data2, n_features):
         return 1.0 - (dot_product / (norm1 * norm2))
 
 
-@numba.njit(cache=True)
+@numba.njit()
 def sparse_hellinger(ind1, data1, ind2, data2):
     aux_inds, aux_data = sparse_mul(ind1, data1, ind2, data2)
     result = 0.0
@@ -618,7 +782,6 @@ def sparse_hellinger(ind1, data1, ind2, data2):
         "dim": numba.types.intp,
         "i": numba.types.uint16,
     },
-    cache=True,
 )
 def sparse_alternative_hellinger(ind1, data1, ind2, data2):
     aux_inds, aux_data = sparse_mul(ind1, data1, ind2, data2)
@@ -649,7 +812,7 @@ def sparse_correct_alternative_hellinger(d):
         return np.sqrt(1.0 - pow(2.0, -d))
 
 
-@numba.njit(cache=True)
+@numba.njit()
 def dummy_ground_metric(x, y):
     return np.float32(not x == y)
 
@@ -690,7 +853,7 @@ def create_ground_metric(ground_vectors, metric):
     return ground_metric
 
 
-@numba.njit(cache=True)
+@numba.njit()
 def sparse_kantorovich(ind1, data1, ind2, data2, ground_metric=dummy_ground_metric):
 
     cost_matrix = np.empty((ind1.shape[0], ind2.shape[0]))
@@ -701,7 +864,83 @@ def sparse_kantorovich(ind1, data1, ind2, data2, ground_metric=dummy_ground_metr
     return kantorovich(data1, data2, cost_matrix)
 
 
-@numba.njit(parallel=True, cache=True)
+@numba.njit()
+def sparse_wasserstein_1d(ind1, data1, ind2, data2, p=1):
+    result = 0.0
+    old_ind = 0
+    delta = 0.0
+    i1 = 0
+    i2 = 0
+    cdf1 = 0.0
+    cdf2 = 0.0
+    l1_norm_x = np.sum(data1)
+    l1_norm_y = np.sum(data2)
+
+    norm = lambda x, p: np.power(np.abs(x), p)
+
+    # pass through both index lists
+    while i1 < ind1.shape[0] and i2 < ind2.shape[0]:
+        j1 = ind1[i1]
+        j2 = ind2[i2]
+
+        if j1 == j2:
+            result += delta * (j1 - old_ind)
+            cdf1 += data1[i1] / l1_norm_x
+            cdf2 += data2[i2] / l1_norm_y
+            delta = norm(cdf1 - cdf2, p)
+            old_ind = j1
+            i1 += 1
+            i2 += 1
+        elif j1 < j2:
+            result += delta * (j1 - old_ind)
+            cdf1 += data1[i1] / l1_norm_x
+            delta = norm(cdf1 - cdf2, p)
+            old_ind = j1
+            i1 += 1
+        else:
+            result += delta * (j2 - old_ind)
+            cdf2 += data2[i2] / l1_norm_y
+            delta = norm(cdf1 - cdf2, p)
+            old_ind = j2
+            i2 += 1
+            # pass over the tails
+    while i1 < ind1.shape[0]:
+        j1 = ind1[i1]
+        result += delta * (j1 - old_ind)
+        cdf1 += data1[i1] / l1_norm_x
+        delta = norm(cdf1 - cdf2, p)
+        old_ind = j1
+        i1 += 1
+
+    while i2 < ind2.shape[0]:
+        j2 = ind2[i2]
+        result += delta * (j2 - old_ind)
+        cdf2 += data2[i2] / l1_norm_y
+        delta = norm(cdf1 - cdf2, p)
+        old_ind = j2
+        i2 += 1
+
+    return np.power(result, 1.0 / p)
+
+
+# Because of the EPS values and the need to normalize after adding them (and then average those for jensen_shannon)
+# it seems like we might as well just take the dense union (dense vectors supported on the union of indices)
+# and call the dense distance functions
+
+
+@numba.njit()
+def sparse_jensen_shannon_divergence(ind1, data1, ind2, data2):
+    dense_data1, dense_data2 = dense_union(ind1, data1, ind2, data2)
+    return jensen_shannon_divergence(dense_data1, dense_data2)
+
+
+@numba.njit()
+def sparse_symmetric_kl_divergence(ind1, data1, ind2, data2):
+    dense_data1, dense_data2 = dense_union(ind1, data1, ind2, data2)
+    return symmetric_kl_divergence(dense_data1, dense_data2)
+
+
+@numba.njit(parallel=True, cache=False)
 def diversify(
     indices,
     distances,
@@ -756,7 +995,7 @@ def diversify(
     return indices, distances
 
 
-@numba.njit(parallel=True, cache=True)
+@numba.njit(parallel=True, cache=False)
 def diversify_csr(
     graph_indptr,
     graph_indices,
@@ -826,8 +1065,6 @@ sparse_named_distances = {
     "minkowski": sparse_minkowski,
     # Other distances
     "canberra": sparse_canberra,
-    "kantorovich": sparse_kantorovich,
-    "wasserstein": sparse_kantorovich,
     "braycurtis": sparse_bray_curtis,
     # Binary distances
     "hamming": sparse_hamming,
@@ -842,7 +1079,19 @@ sparse_named_distances = {
     # Angular distances
     "cosine": sparse_cosine,
     "correlation": sparse_correlation,
+    # Distribution distances
+    "kantorovich": sparse_kantorovich,
+    "wasserstein": sparse_kantorovich,
+    "wasserstein_1d": sparse_wasserstein_1d,
+    "wasserstein-1d": sparse_wasserstein_1d,
+    "kantorovich-1d": sparse_wasserstein_1d,
+    "kantorovich-1d": sparse_wasserstein_1d,
     "hellinger": sparse_hellinger,
+    "jensen-shannon": sparse_jensen_shannon_divergence,
+    "jensen_shannon": sparse_jensen_shannon_divergence,
+    "symmetric-kl": sparse_symmetric_kl_divergence,
+    "symmetric_kl": sparse_symmetric_kl_divergence,
+    "symmetric_kullback_liebler": sparse_symmetric_kl_divergence,
 }
 
 sparse_need_n_features = (
