@@ -5,8 +5,6 @@
 import time
 
 import numba
-from numba.core import types
-import numba.experimental.structref as structref
 import numpy as np
 
 
@@ -69,7 +67,7 @@ def tau_rand(state):
     locals={
         "dim": numba.types.intp,
         "i": numba.types.uint32,
-        "result": numba.types.float32,
+        # "result": numba.types.float32, # This provides speed, but causes errors in corner cases
     },
     fastmath=True,
     cache=True,
@@ -129,45 +127,6 @@ def rejection_sample(n_samples, pool_size, rng_state):
     return result
 
 
-@structref.register
-class HeapType(types.StructRef):
-    pass
-
-
-class Heap(structref.StructRefProxy):
-    @property
-    def indices(self):
-        return Heap_get_indices(self)
-
-    @property
-    def distances(self):
-        return Heap_get_distances(self)
-
-    @property
-    def flags(self):
-        return Heap_get_flags(self)
-
-
-@numba.njit(cache=True)
-def Heap_get_flags(self):
-    return self.flags
-
-
-@numba.njit(cache=True)
-def Heap_get_distances(self):
-    return self.distances
-
-
-@numba.njit(cache=True)
-def Heap_get_indices(self):
-    return self.indices
-
-
-structref.define_proxy(Heap, HeapType, ["indices", "distances", "flags"])
-
-# Heap = namedtuple("Heap", ("indices", "distances", "flags"))
-
-
 @numba.njit(cache=True)
 def make_heap(n_points, size):
     """Constructor for the numba enabled heap objects. The heaps are used
@@ -192,203 +151,15 @@ def make_heap(n_points, size):
     heap: An ndarray suitable for passing to other numba enabled heap functions.
     """
     indices = np.full((int(n_points), int(size)), -1, dtype=np.int32)
-    distances = np.full((int(n_points), int(size)), np.infty, dtype=np.float32)
+    distances = np.full((int(n_points), int(size)), np.inf, dtype=np.float32)
     flags = np.zeros((int(n_points), int(size)), dtype=np.uint8)
     result = (indices, distances, flags)
 
     return result
 
 
-@numba.jit(
-    locals={
-        "indices": numba.types.int32[::1],
-        "weights": numba.types.float32[::1],
-        "is_new": numba.types.uint8[::1],
-        "i": numba.types.uint16,
-        "ic1": numba.types.uint16,
-        "ic2": numba.types.uint16,
-        "i_swap": numba.types.uint16,
-        "heap_size": numba.types.uint16,
-    },
-    cache=True,
-)
-def heap_push(heap, row, weight, index, flag):
-    """Push a new element onto the heap. The heap stores potential neighbors
-    for each graph_data point. The ``row`` parameter determines which graph_data point we
-    are addressing, the ``weight`` determines the distance (for heap sorting),
-    the ``index`` is the element to add, and the flag determines whether this
-    is to be considered a new addition.
-
-    Parameters
-    ----------
-    heap: ndarray generated by ``make_heap``
-        The heap object to push into
-
-    row: int
-        Which actual heap within the heap object to push to
-
-    weight: float
-        The priority value of the element to push onto the heap
-
-    index: int
-        The actual value to be pushed
-
-    flag: int
-        Whether to flag the newly added element or not.
-
-    Returns
-    -------
-    success: The number of new elements successfully pushed into the heap.
-    """
-    row = np.int32(row)
-    weight = np.float32(weight)
-    index = np.int32(index)
-    flag = np.uint8(flag)
-
-    indices = heap[0][row]
-    weights = heap[1][row]
-    is_new = heap[2][row]
-
-    if weight >= weights[0]:
-        return 0
-
-    # break if we already have this element.
-    for i in range(indices.shape[0]):
-        if index == indices[i]:
-            return 0
-
-    # insert val at position zero
-    weights[0] = weight
-    indices[0] = index
-    is_new[0] = flag
-
-    # descend the heap, swapping values until the max heap criterion is met
-    i = 0
-    while True:
-        ic1 = 2 * i + 1
-        ic2 = ic1 + 1
-
-        if ic1 >= indices.shape[0]:
-            break
-        elif ic2 >= indices.shape[0]:
-            if weights[ic1] > weight:
-                i_swap = ic1
-            else:
-                break
-        elif weights[ic1] >= weights[ic2]:
-            if weight < weights[ic1]:
-                i_swap = ic1
-            else:
-                break
-        else:
-            if weight < weights[ic2]:
-                i_swap = ic2
-            else:
-                break
-
-        weights[i] = weights[i_swap]
-        indices[i] = indices[i_swap]
-        is_new[i] = is_new[i_swap]
-
-        i = i_swap
-
-    weights[i] = weight
-    indices[i] = index
-    is_new[i] = flag
-
-    return 1
-
-
-@numba.jit(
-    locals={
-        "indices": numba.types.int32[::1],
-        "weights": numba.types.float32[::1],
-        "is_new": numba.types.uint8[::1],
-        "i": numba.types.uint16,
-        "ic1": numba.types.uint16,
-        "ic2": numba.types.uint16,
-        "i_swap": numba.types.uint16,
-        "heap_size": numba.types.uint16,
-    },
-    cache=True,
-)
-def unchecked_heap_push(heap, row, weight, index, flag):
-    """Push a new element onto the heap. The heap stores potential neighbors
-    for each graph_data point. The ``row`` parameter determines which graph_data point we
-    are addressing, the ``weight`` determines the distance (for heap sorting),
-    the ``index`` is the element to add, and the flag determines whether this
-    is to be considered a new addition.
-
-    Parameters
-    ----------
-    heap: ndarray generated by ``make_heap``
-        The heap object to push into
-
-    row: int
-        Which actual heap within the heap object to push to
-
-    weight: float
-        The priority value of the element to push onto the heap
-
-    index: int
-        The actual value to be pushed
-
-    flag: int
-        Whether to flag the newly added element or not.
-
-    Returns
-    -------
-    success: The number of new elements successfully pushed into the heap.
-    """
-    if weight >= heap[1][row, 0]:
-        return 0
-
-    indices = heap[0][row]
-    weights = heap[1][row]
-    is_new = heap[2][row]
-
-    # insert val at position zero
-    weights[0] = weight
-    indices[0] = index
-    is_new[0] = flag
-
-    heap_size = indices.shape[0]
-
-    # descend the heap, swapping values until the max heap criterion is met
-    i = 0
-    while True:
-        ic1 = 2 * i + 1
-        ic2 = ic1 + 1
-
-        if ic1 >= heap_size:
-            break
-        elif ic2 >= heap_size:
-            if weights[ic1] > weight:
-                i_swap = ic1
-            else:
-                break
-        elif weights[ic1] >= weights[ic2]:
-            if weight < weights[ic1]:
-                i_swap = ic1
-            else:
-                break
-        else:
-            if weight < weights[ic2]:
-                i_swap = ic2
-            else:
-                break
-
-        weights[i] = weights[i_swap]
-        indices[i] = indices[i_swap]
-        is_new[i] = is_new[i_swap]
-
-        i = i_swap
-
-    weights[i] = weight
-    indices[i] = index
-    is_new[i] = flag
-
-    return 1
+# Sentinel value for empty/uninitialized graphs
+EMPTY_GRAPH = make_heap(1, 1)
 
 
 @numba.njit(cache=True)
@@ -415,90 +186,39 @@ def siftdown(heap1, heap2, elt):
             elt = swap
 
 
-@numba.njit(cache=True)
-def deheap_sort(heap):
-    """Given an array of heaps (of graph_indices and weights), unpack the heap
-    out to give and array of sorted lists of graph_indices and weights by increasing
-    weight. This is effectively just the second half of heap sort (the first
-    half not being required since we already have the graph_data in a heap).
+@numba.njit(parallel=True, cache=False)
+def deheap_sort(indices, distances):
+    """Given two arrays representing a heap (indices and distances), reorder the
+     arrays by increasing distance. This is effectively just the second half of
+     heap sort (the first half not being required since we already have the
+     graph_data in a heap).
+
+     Note that this is done in-place.
 
     Parameters
     ----------
-    heap : array of shape (3, n_samples, n_neighbors)
-        The heap to turn into sorted lists.
+    indices : array of shape (n_samples, n_neighbors)
+        The graph indices to sort by distance.
+    distances : array of shape (n_samples, n_neighbors)
+        The corresponding edge distance.
 
     Returns
     -------
-    graph_indices, weights: arrays of shape (n_samples, n_neighbors)
-        The graph_indices and weights sorted by increasing weight.
+    indices, distances: arrays of shape (n_samples, n_neighbors)
+        The indices and distances sorted by increasing distance.
     """
-    indices = heap[0]
-    weights = heap[1]
+    for i in numba.prange(indices.shape[0]):
+        # starting from the end of the array and moving back
+        for j in range(indices.shape[1] - 1, 0, -1):
+            indices[i, 0], indices[i, j] = indices[i, j], indices[i, 0]
+            distances[i, 0], distances[i, j] = distances[i, j], distances[i, 0]
 
-    for i in range(indices.shape[0]):
+            siftdown(distances[i, :j], indices[i, :j], 0)
 
-        ind_heap = indices[i]
-        dist_heap = weights[i]
-
-        for j in range(ind_heap.shape[0] - 1):
-            ind_heap[0], ind_heap[ind_heap.shape[0] - j - 1] = (
-                ind_heap[ind_heap.shape[0] - j - 1],
-                ind_heap[0],
-            )
-            dist_heap[0], dist_heap[dist_heap.shape[0] - j - 1] = (
-                dist_heap[dist_heap.shape[0] - j - 1],
-                dist_heap[0],
-            )
-
-            siftdown(
-                dist_heap[: dist_heap.shape[0] - j - 1],
-                ind_heap[: ind_heap.shape[0] - j - 1],
-                0,
-            )
-
-    return indices.astype(np.int64), weights
+    return indices, distances
 
 
-# @numba.njit()
-# def smallest_flagged(heap, row):
-#     """Search the heap for the smallest element that is
-#     still flagged.
-#
-#     Parameters
-#     ----------
-#     heap: array of shape (3, n_samples, n_neighbors)
-#         The heaps to search
-#
-#     row: int
-#         Which of the heaps to search
-#
-#     Returns
-#     -------
-#     index: int
-#         The index of the smallest flagged element
-#         of the ``row``th heap, or -1 if no flagged
-#         elements remain in the heap.
-#     """
-#     ind = heap[0][row]
-#     dist = heap[1][row]
-#     flag = heap[2][row]
-#
-#     min_dist = np.inf
-#     result_index = -1
-#
-#     for i in range(ind.shape[0]):
-#         if flag[i] == 1 and dist[i] < min_dist:
-#             min_dist = dist[i]
-#             result_index = i
-#
-#     if result_index >= 0:
-#         flag[result_index] = 0.0
-#         return int(ind[result_index])
-#     else:
-#         return -1
-
-
-@numba.njit(parallel=True, locals={"idx": numba.types.int64}, cache=True)
+@numba.njit(parallel=True, locals={"idx": numba.types.int64}, cache=False)
 def new_build_candidates(current_graph, max_candidates, rng_state, n_threads):
     """Build a heap of candidate neighbors for nearest neighbor descent. For
     each vertex the candidate neighbors are any current neighbors, and any
@@ -536,42 +256,54 @@ def new_build_candidates(current_graph, max_candidates, rng_state, n_threads):
         (n_vertices, max_candidates), np.inf, dtype=np.float32
     )
 
+    block_size = n_vertices // n_threads + 1
+
     for n in numba.prange(n_threads):
         local_rng_state = rng_state + n
+        block_start = n * block_size
+        block_end = min(block_start + block_size, n_vertices)
+
         for i in range(n_vertices):
             for j in range(n_neighbors):
                 idx = current_indices[i, j]
-                isn = current_flags[i, j]
 
-                if idx < 0:
-                    continue
+                if idx >= 0 and (
+                    (i >= block_start and i < block_end)
+                    or (idx >= block_start and idx < block_end)
+                ):
+                    isn = current_flags[i, j]
+                    d = tau_rand(local_rng_state)
 
-                d = tau_rand(local_rng_state)
-
-                if isn:
-                    if i % n_threads == n:
-                        checked_heap_push(
-                            new_candidate_priority[i], new_candidate_indices[i], d, idx
-                        )
-                    if idx % n_threads == n:
-                        checked_heap_push(
-                            new_candidate_priority[idx],
-                            new_candidate_indices[idx],
-                            d,
-                            i,
-                        )
-                else:
-                    if i % n_threads == n:
-                        checked_heap_push(
-                            old_candidate_priority[i], old_candidate_indices[i], d, idx
-                        )
-                    if idx % n_threads == n:
-                        checked_heap_push(
-                            old_candidate_priority[idx],
-                            old_candidate_indices[idx],
-                            d,
-                            i,
-                        )
+                    if isn:
+                        if i >= block_start and i < block_end:
+                            checked_heap_push(
+                                new_candidate_priority[i],
+                                new_candidate_indices[i],
+                                d,
+                                idx,
+                            )
+                        if idx >= block_start and idx < block_end:
+                            checked_heap_push(
+                                new_candidate_priority[idx],
+                                new_candidate_indices[idx],
+                                d,
+                                i,
+                            )
+                    else:
+                        if i >= block_start and i < block_end:
+                            checked_heap_push(
+                                old_candidate_priority[i],
+                                old_candidate_indices[i],
+                                d,
+                                idx,
+                            )
+                        if idx >= block_start and idx < block_end:
+                            checked_heap_push(
+                                old_candidate_priority[idx],
+                                old_candidate_indices[idx],
+                                d,
+                                i,
+                            )
 
     indices = current_graph[0]
     flags = current_graph[2]
@@ -603,6 +335,20 @@ def mark_visited(table, candidate):
     return
 
 
+@numba.njit("b1(u1[::1],i4)", cache=True)
+def check_and_mark_visited(table, candidate):
+    """Check if candidate was visited and mark it as visited in one operation.
+    
+    Returns True if the candidate was already visited, False otherwise.
+    More efficient than separate has_been_visited + mark_visited calls.
+    """
+    loc = candidate >> 3
+    mask = numba.uint8(1 << (candidate & 7))
+    was_visited = table[loc] & mask
+    table[loc] |= mask
+    return was_visited
+
+
 @numba.njit(
     "i4(f4[::1],i4[::1],f4,i4)",
     fastmath=True,
@@ -722,66 +468,6 @@ def checked_heap_push(priorities, indices, p, n):
     return 1
 
 
-@numba.njit(
-    "i4(f4[::1],i4[::1],u1[::1],f4,i4,u1)",
-    fastmath=True,
-    locals={
-        "size": numba.types.intp,
-        "i": numba.types.uint16,
-        "ic1": numba.types.uint16,
-        "ic2": numba.types.uint16,
-        "i_swap": numba.types.uint16,
-    },
-    cache=True,
-)
-def flagged_heap_push(priorities, indices, flags, p, n, f):
-    if p >= priorities[0]:
-        return 0
-
-    size = priorities.shape[0]
-
-    # insert val at position zero
-    priorities[0] = p
-    indices[0] = n
-    flags[0] = f
-
-    # descend the heap, swapping values until the max heap criterion is met
-    i = 0
-    while True:
-        ic1 = 2 * i + 1
-        ic2 = ic1 + 1
-
-        if ic1 >= size:
-            break
-        elif ic2 >= size:
-            if priorities[ic1] > p:
-                i_swap = ic1
-            else:
-                break
-        elif priorities[ic1] >= priorities[ic2]:
-            if p < priorities[ic1]:
-                i_swap = ic1
-            else:
-                break
-        else:
-            if p < priorities[ic2]:
-                i_swap = ic2
-            else:
-                break
-
-        priorities[i] = priorities[i_swap]
-        indices[i] = indices[i_swap]
-        flags[i] = flags[i_swap]
-
-        i = i_swap
-
-    priorities[i] = p
-    indices[i] = n
-    flags[i] = f
-
-    return 1
-
-
 @numba.njit(
     "i4(f4[::1],i4[::1],u1[::1],f4,i4,u1)",
     fastmath=True,
@@ -849,42 +535,196 @@ def checked_flagged_heap_push(priorities, indices, flags, p, n, f):
 
 @numba.njit(
     parallel=True,
+    locals={
+        "dist_thresh_p": numba.float32,
+        "dist_thresh_q": numba.float32,
+        "p": numba.int32,
+        "q": numba.int32,
+        "d": numba.float32,
+        "max_updates": numba.int32,
+        "max_threshold": numba.float32,
+    },
+    cache=False,
+    fastmath=True,
+)
+def generate_graph_update_array(
+    update_array,
+    n_updates_per_thread,
+    new_candidate_block,
+    old_candidate_block,
+    dist_thresholds,
+    data,
+    dist,
+    n_threads,
+):
+    """Generate graph updates into a pre-allocated array.
+
+    This is more efficient than generating lists of tuples because:
+    1. No dynamic memory allocation during the parallel loop
+    2. Better cache locality with contiguous array storage
+    3. Each thread writes to its own section of the array
+
+    Parameters
+    ----------
+    update_array : ndarray of shape (n_threads, max_updates_per_thread, 3)
+        Pre-allocated array to store updates. Each row stores (p, q, d).
+
+    n_updates_per_thread : ndarray of shape (n_threads,)
+        Output array to store the number of updates generated by each thread.
+
+    new_candidate_block : ndarray of shape (block_size, max_candidates)
+        New candidate indices for this block.
+
+    old_candidate_block : ndarray of shape (block_size, max_candidates)
+        Old candidate indices for this block.
+
+    dist_thresholds : ndarray of shape (n_vertices,)
+        Current distance thresholds (max heap distance) for each vertex.
+
+    data : ndarray of shape (n_vertices, n_features)
+        The data points.
+
+    dist : callable
+        Distance function.
+
+    n_threads : int
+        Number of threads to use.
+    """
+    block_size = new_candidate_block.shape[0]
+    max_new_candidates = new_candidate_block.shape[1]
+    max_old_candidates = old_candidate_block.shape[1]
+    rows_per_thread = (block_size // n_threads) + 1
+
+    for t in numba.prange(n_threads):
+        idx = 0
+        max_updates = update_array.shape[1]
+
+        for r in range(rows_per_thread):
+            i = t * rows_per_thread + r
+            if i >= block_size or idx >= max_updates:
+                break
+
+            for j in range(max_new_candidates):
+                if idx >= max_updates:
+                    break
+
+                p = new_candidate_block[i, j]
+                if p < 0:
+                    continue
+
+                data_p = data[p]
+                dist_thresh_p = dist_thresholds[p]
+
+                # Compare with other new candidates (start at j to match original behavior)
+                for k in range(j, max_new_candidates):
+                    if idx >= max_updates:
+                        break
+
+                    q = new_candidate_block[i, k]
+                    if q < 0:
+                        continue
+
+                    d = dist(data_p, data[q])
+
+                    # Use max for better branch prediction than OR condition
+                    dist_thresh_q = dist_thresholds[q]
+                    max_threshold = max(dist_thresh_p, dist_thresh_q)
+
+                    if d <= max_threshold:
+                        update_array[t, idx, 0] = p
+                        update_array[t, idx, 1] = q
+                        update_array[t, idx, 2] = d
+                        idx += 1
+
+                # Compare with old candidates
+                for k in range(max_old_candidates):
+                    if idx >= max_updates:
+                        break
+
+                    q = old_candidate_block[i, k]
+                    if q < 0:
+                        continue
+
+                    d = dist(data_p, data[q])
+                    dist_thresh_q = dist_thresholds[q]
+                    max_threshold = max(dist_thresh_p, dist_thresh_q)
+
+                    if d <= max_threshold:
+                        update_array[t, idx, 0] = p
+                        update_array[t, idx, 1] = q
+                        update_array[t, idx, 2] = d
+                        idx += 1
+
+        n_updates_per_thread[t] = idx
+
+
+@numba.njit(
+    parallel=True,
+    cache=True,
     locals={
         "p": numba.int32,
         "q": numba.int32,
         "d": numba.float32,
         "added": numba.uint8,
         "n": numba.uint32,
-        "i": numba.uint32,
+        "t": numba.uint32,
         "j": numba.uint32,
     },
-    cache=True,
 )
-def apply_graph_updates_low_memory(current_graph, updates, n_threads):
+def apply_graph_update_array(
+    current_graph, update_array, n_updates_per_thread, n_threads
+):
+    """Apply graph updates from a pre-allocated array.
 
+    Uses block-based processing where each thread only updates vertices
+    in its assigned block, avoiding the need for duplicate checking.
+
+    Parameters
+    ----------
+    current_graph : tuple of (indices, distances, flags)
+        The current nearest neighbor graph heap.
+
+    update_array : ndarray of shape (n_threads, max_updates_per_thread, 3)
+        Array of updates where each row is (p, q, d).
+
+    n_updates_per_thread : ndarray of shape (n_threads,)
+        Number of valid updates from each generating thread.
+
+    n_threads : int
+        Number of threads.
+
+    Returns
+    -------
+    n_changes : int
+        Total number of updates that modified the graph.
+    """
     n_changes = 0
     priorities = current_graph[1]
     indices = current_graph[0]
     flags = current_graph[2]
-    # n_threads = numba.get_num_threads()
+
+    n_vertices = priorities.shape[0]
+    vertex_block_size = n_vertices // n_threads + 1
 
     for n in numba.prange(n_threads):
-        for i in range(len(updates)):
-            for j in range(len(updates[i])):
-                p, q, d = updates[i][j]
+        block_start = n * vertex_block_size
+        block_end = min(block_start + vertex_block_size, n_vertices)
 
-                if p == -1 or q == -1:
-                    continue
+        # Each thread scans all updates but only applies those
+        # where p or q falls in its block
+        for t in range(n_threads):
+            for j in range(n_updates_per_thread[t]):
+                p = np.int32(update_array[t, j, 0])
+                q = np.int32(update_array[t, j, 1])
+                d = np.float32(update_array[t, j, 2])
 
-                if p % n_threads == n:
-                    # added = heap_push(current_graph, p, d, q, 1)
+                if p >= block_start and p < block_end:
                     added = checked_flagged_heap_push(
                         priorities[p], indices[p], flags[p], d, q, 1
                     )
                     n_changes += added
 
-                if q % n_threads == n:
-                    # added = heap_push(current_graph, q, d, p, 1)
+                if q >= block_start and q < block_end:
                     added = checked_flagged_heap_push(
                         priorities[q], indices[q], flags[q], d, p, 1
                     )
@@ -893,58 +733,107 @@ def apply_graph_updates_low_memory(current_graph, updates, n_threads):
     return n_changes
 
 
-@numba.njit(locals={"p": numba.types.int64, "q": numba.types.int64}, cache=True)
-def apply_graph_updates_high_memory(current_graph, updates, in_graph):
+@numba.njit(
+    parallel=True,
+    cache=False,
+    fastmath=True,
+    locals={
+        "dist_thresh_p": numba.float32,
+        "dist_thresh_q": numba.float32,
+        "p": numba.int32,
+        "q": numba.int32,
+        "d": numba.float32,
+        "max_updates": numba.int32,
+        "max_threshold": numba.float32,
+    },
+)
+def sparse_generate_graph_update_array(
+    update_array,
+    n_updates_per_thread,
+    new_candidate_block,
+    old_candidate_block,
+    dist_thresholds,
+    inds,
+    indptr,
+    data,
+    dist,
+    n_threads,
+):
+    """Generate graph updates for sparse data into a pre-allocated array."""
+    block_size = new_candidate_block.shape[0]
+    max_new_candidates = new_candidate_block.shape[1]
+    max_old_candidates = old_candidate_block.shape[1]
+    rows_per_thread = (block_size // n_threads) + 1
 
-    n_changes = 0
+    for t in numba.prange(n_threads):
+        idx = 0
+        max_updates = update_array.shape[1]
 
-    for i in range(len(updates)):
-        for j in range(len(updates[i])):
-            p, q, d = updates[i][j]
+        for r in range(rows_per_thread):
+            i = t * rows_per_thread + r
+            if i >= block_size or idx >= max_updates:
+                break
 
-            if p == -1 or q == -1:
-                continue
+            for j in range(max_new_candidates):
+                if idx >= max_updates:
+                    break
 
-            if q in in_graph[p] and p in in_graph[q]:
-                continue
-            elif q in in_graph[p]:
-                pass
-            else:
-                # added = unchecked_heap_push(current_graph, p, d, q, 1)
-                added = flagged_heap_push(
-                    current_graph[1][p],
-                    current_graph[0][p],
-                    current_graph[2][p],
-                    d,
-                    q,
-                    1,
-                )
+                p = new_candidate_block[i, j]
+                if p < 0:
+                    continue
 
-                if added > 0:
-                    in_graph[p].add(q)
-                    n_changes += added
+                from_inds = inds[indptr[p] : indptr[p + 1]]
+                from_data = data[indptr[p] : indptr[p + 1]]
+                dist_thresh_p = dist_thresholds[p]
 
-            if p == q or p in in_graph[q]:
-                pass
-            else:
-                # added = unchecked_heap_push(current_graph, q, d, p, 1)
-                added = flagged_heap_push(
-                    current_graph[1][p],
-                    current_graph[0][p],
-                    current_graph[2][p],
-                    d,
-                    q,
-                    1,
-                )
+                # Compare with other new candidates (start at j to match original)
+                for k in range(j, max_new_candidates):
+                    if idx >= max_updates:
+                        break
 
-                if added > 0:
-                    in_graph[q].add(p)
-                    n_changes += added
+                    q = new_candidate_block[i, k]
+                    if q < 0:
+                        continue
 
-    return n_changes
+                    to_inds = inds[indptr[q] : indptr[q + 1]]
+                    to_data = data[indptr[q] : indptr[q + 1]]
+                    d = dist(from_inds, from_data, to_inds, to_data)
+
+                    dist_thresh_q = dist_thresholds[q]
+                    max_threshold = max(dist_thresh_p, dist_thresh_q)
+
+                    if d <= max_threshold:
+                        update_array[t, idx, 0] = p
+                        update_array[t, idx, 1] = q
+                        update_array[t, idx, 2] = d
+                        idx += 1
+
+                # Compare with old candidates
+                for k in range(max_old_candidates):
+                    if idx >= max_updates:
+                        break
+
+                    q = old_candidate_block[i, k]
+                    if q < 0:
+                        continue
+
+                    to_inds = inds[indptr[q] : indptr[q + 1]]
+                    to_data = data[indptr[q] : indptr[q + 1]]
+                    d = dist(from_inds, from_data, to_inds, to_data)
+
+                    dist_thresh_q = dist_thresholds[q]
+                    max_threshold = max(dist_thresh_p, dist_thresh_q)
+
+                    if d <= max_threshold:
+                        update_array[t, idx, 0] = p
+                        update_array[t, idx, 1] = q
+                        update_array[t, idx, 2] = d
+                        idx += 1
+
+        n_updates_per_thread[t] = idx
 
 
-@numba.njit(cache=True)
+@numba.njit(cache=False)
 def initalize_heap_from_graph_indices(heap, graph_indices, data, metric):
 
     for i in range(graph_indices.shape[0]):
@@ -952,12 +841,26 @@ def initalize_heap_from_graph_indices(heap, graph_indices, data, metric):
             j = graph_indices[i, idx]
             if j >= 0:
                 d = metric(data[i], data[j])
-                flagged_heap_push(heap[1][i], heap[0][i], heap[2][i], d, j, 1)
+                checked_flagged_heap_push(heap[1][i], heap[0][i], heap[2][i], d, j, 1)
 
     return heap
 
 
-@numba.njit(parallel=True, cache=True)
+@numba.njit(cache=True)
+def initalize_heap_from_graph_indices_and_distances(
+    heap, graph_indices, graph_distances
+):
+    for i in range(graph_indices.shape[0]):
+        for idx in range(graph_indices.shape[1]):
+            j = graph_indices[i, idx]
+            if j >= 0:
+                d = graph_distances[i, idx]
+                checked_flagged_heap_push(heap[1][i], heap[0][i], heap[2][i], d, j, 1)
+
+    return heap
+
+
+@numba.njit(parallel=True, cache=False)
 def sparse_initalize_heap_from_graph_indices(
     heap, graph_indices, data_indptr, data_indices, data_vals, metric
 ):
@@ -970,8 +873,7 @@ def sparse_initalize_heap_from_graph_indices(
             ind2 = data_indices[data_indptr[j] : data_indptr[j + 1]]
             data2 = data_vals[data_indptr[j] : data_indptr[j + 1]]
             d = metric(ind1, data1, ind2, data2)
-            # unchecked_heap_push(heap, i, d, j, 1)
-            flagged_heap_push(heap[0][i], heap[1][i], heap[2][i], j, d, 1)
+            checked_flagged_heap_push(heap[1][i], heap[0][i], heap[2][i], d, j, 1)
 
     return heap
 
